unit CheckListBx;
(*


   ########### NOTES ######################################################################
   ##                                                                                    ##
   ##  Feel free to hack or use in any way                                              ##
   ##                                                                                    ##
   ##  Originated by Peter Fox                peter@eminent.demon.co.uk                 ##
   ##                                                                                    ##
   ##  Source is freely available at          www.eminent.demon.co.uk/delphi.htm        ##
   ##                                                                                    ##
   ##  Please drop me a line if you find it useful                                      ##
   ##                                                                                    ##
   ##  Comments,corrections and additions are all welcome                               ##
   ##                                                                                    ##
   ##  This version was put to bed on 27th Dec 2001                                     ##
   ##                                                                                    ##
   ########################################################################################


  A dialog check-list which can operate in three modes
   Pick one     - Single selection triggers return
   Pick some    - none to all items can be selected - No automatic return
   Check all    - normally waits for all items to be ticked before returning true

  The programmer sets title and message/instructions
   The box will resize to fit the message
   Message text can be broken onto multiple lines (there is no word wrap) with vertical bars
  eg ('OK to print','Have you filled in| *all* |the necessary data',...)

  The item list is set as strings split with vertical bars.
  eg 'North|south|East|West'.
  Prefix an item with a '+' to start off with it checked.
  eg  '+Usual option|Another possibility'

  Different styles of tick-box are provided for the different functions
   Pick-one items are shown as circles (AKA radio buttons)
   Pick some are shown as tick and cross
   Check-all are shown as boxes and ticked boxes

  Buttons
   Help button optional
   OK/Cancel buttons are optional where relavent

  Controls
   Esc cancels
   Return accepts (but only appropriate for pick some)
   Mouse click toggles tick
   Space toggles tick and moves down to next one
   Up and down arrows move highlight
   Initial letters will move to next item
   Unique initial letters will (un)tick appropriate item

  Varying return types
   Pick-one returns the index of the item picked or -1 if cancelled
   Pick-some returns a set of integers listing the ticked items
   Check-all returns a boolean true only if all are ticked

  Examples
  --------------------------------------------------------------------------
  | if Pickone('Sex','What is the gender|of '+babyname,'Boy|Girl') < 0 then
  |   showmessage('Please pay attention at the back!);
  --------------------------------------------------------------------------
  | var
  |   R : TcbxResult;
  | begin
  |   R := PickSome('Pets','What animals do you have','Cat|Dog|Rabbit|Other',true);
  |   if R=[] then showmessage('You haven''t got any pets');
  |   if 2 in R then showmessage('Woof Woof');
  --------------------------------------------------------------------------
  | if CheckAll('Off we go!','Last minute checks','Turn off gas|Passport|Tickets') then
  |   showmessage('Have a nice holiday');
  --------------------------------------------------------------------------

   
                                                                                         
     It is worth thinking for a moment about the idiom of check lists in the context     
     of a set of tasks that the operator is being led through.  The formal check-list    
     is a compound reminder which helps with diligence.  The idea of a dialog which      
     offers options is to focus on a particular decision at a particular point of the    
     process.  So this can lead to a style of operation which is lots of small           
     decisions or inputs in a structured sequence. (As opposed to the "fill in all the   
     items on this form then press Accept") I'm not advocating one style or the other    
     but it may be worth while to consider where the 'wizard' approach may me more       
     effective.                                                                          
                                                                                         
     I first used this style (not just with check boxes) where five aspects of a task    
     were completed in turn (pick this,pick that,input quantity etc) followed by a       
     final confirmation check-box where the selections could be reviewed just to make    
     sure.  This was a medical application but there are plenty of cases where a double  
     check is worthwhile.                                                                
                                                                                         
     Following on from the idea of the final check-box is a 1-in-a-100 deliberate        
     insertion of an error to keep the operator on their toes.                           
                                                                                         
   




*)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls;

type
  TcbxMode = (cbxmPickOne,cbxmPickSome,cbxmCheckAll);
  TcbxInt  = 0..65;   // 65=out of range marker  (NoIX const)
  TcbxResult = set of Tcbxint;

  TfCheckListBx = class(TForm)
    BuOK: TBitBtn;
    buCancel: TBitBtn;
    BuHelp: TBitBtn;
    ticks: TImage;
    procedure FormPaint(Sender: TObject);
    procedure BuHelpClick(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    fmode    : Tcbxmode;
    fchecked : TCbxresult;
    fmsg     : string;
    fHlpCtxt : integer;
    fItems   : Tstringlist;
    fX0      : integer;        // list left
    fY0      : integer;        // list top
    fLineHt  : integer;        // list line height
    fcurrentIx : TcbxInt;
    f1stChars:string;          // list of first characters of items
    procedure SetCurrentix(const ix:TcbxInt);
    procedure DrawItem(const ix:TCbxInt);
    procedure DrawTick(const ix:TCbxInt);
    procedure Flipcheck;
    function  Checked:TCbxresult;
    function  NotChecked:TCbxresult;
    function  chop(var Mainstring : string; const Chopper : string):string;
    function  Execute(const mode : Tcbxmode;
                     const Title,msg,ItemList : string;
                     const ShowButtons : boolean;
                     const hlpctxt:THelpContext):TcbxResult;
  public
    { Public declarations }
  end;

// PICKONE returns 0...n-1 or -1 if cancelled
function PickOne(const Title,msg,ItemList:string):integer;                                                     overload;
function PickOne(const Title,msg,ItemList:string;const ShowButtons:boolean):integer;                           overload;
function PickOne(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):integer;overload;

// PICKSOME returns a TcbxResult which is a set of integers [0..n-1]
// -1 in the set indicates cancelled
function PickSome(const Title,msg,ItemList:string):TcbxResult;                                                     overload;
function PickSome(const Title,msg,ItemList:string;const ShowButtons:boolean):TcbxResult;                           overload;
function PickSome(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):TcbxResult;overload;

// CHECKALL returns true if all items are checked
function CheckAll(const Title,msg,ItemList:string):boolean;                                                     overload;
function CheckAll(const Title,msg,ItemList:string;const hlpctxt:THelpContext):boolean;                          overload;
function CheckAll(const Title,msg,ItemList:string;const ShowButtons:boolean):boolean;                           overload;
function CheckAll(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):boolean;overload;

// This is the master call
function CheckListBxExecute(const mode : Tcbxmode;
                            const Title,msg,ItemList : string;
                            const ShowButtons : boolean;
                            const hlpctxt:THelpContext):TcbxResult;

var
  fCheckListBx: TfCheckListBx;
const
  noix = 65;  // flag out of range item index

implementation

{$R *.DFM}

// 
//                                                                              
//  implementations of helper calls                                             
//                                                                              
// 



// convert the front-end functions to the master call
function PickOne(const Title,msg,ItemList:string):integer;
begin    result := pickone(title,msg,itemlist,false,0);  end;

function PickOne(const Title,msg,ItemList:string;const ShowButtons:boolean):integer;
begin    result := pickone(title,msg,itemlist,ShowButtons,0);  end;

function PickOne(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):integer;
var
  pl :TcbxResult;
  i  : integer;
begin
  pl := CheckListBxExecute(cbxmPickOne,Title,msg,ItemList,ShowButtons,hlpctxt);
  result := -1;
  if not (pl = []) then begin
    for i := 0 to high(TcbxInt) do begin
      if i in pl then begin
        result := i;
        break;
      end;
    end;
  end;
end;


// PICKSOME returns a TcbxResult which is a set of TcbxInt(integers 0..63) [0..n-1]
// -1 in the set indicates cancelled
function PickSome(const Title,msg,ItemList:string):TcbxResult;
begin    result := PickSome(Title,msg,ItemList,false,0); end;

function PickSome(const Title,msg,ItemList:string;const ShowButtons:boolean):TcbxResult;
begin    result := PickSome(Title,msg,ItemList,ShowButtons,0); end;

function PickSome(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):TcbxResult;
begin
  result := CheckListBxExecute(cbxmPickSome,Title,msg,ItemList,ShowButtons,hlpctxt);
end;

// CHECKALL returns true if all items are checked
function CheckAll(const Title,msg,ItemList:string):boolean;
begin    result := CheckAll(Title,msg,ItemList,false,0);    end;

function CheckAll(const Title,msg,ItemList:string;const hlpctxt:THelpContext):boolean;
begin    result := CheckAll(Title,msg,ItemList,false,hlpctxt);    end;

function CheckAll(const Title,msg,ItemList:string;const ShowButtons:boolean):boolean;
begin    result := CheckAll(Title,msg,ItemList,ShowButtons,0);    end;

function CheckAll(const Title,msg,ItemList:string;const ShowButtons:boolean;const hlpctxt:THelpContext):boolean;
var
  pl :TcbxResult;
begin
  pl := CheckListBxExecute(cbxmCheckAll,Title,msg,ItemList,ShowButtons,hlpctxt);
  result := (pl = []);  // cbxmCheckAll mode returns UNCHECKED items
end;


// This is the master call
function CheckListBxExecute(const mode : Tcbxmode;
                            const Title,msg,ItemList : string;
                            const ShowButtons : boolean;
                            const hlpctxt:THelpContext):TcbxResult;
{ Results are returned in the set of TcbxInt(integers 0..63) in the function result
  pickone  : Set is empty - cancelled or not selected
             Otherwise picked item is low(result)
  picksome : Set may be empty!
             Otherwise picked items are itemised in result
  checkall : result is *UNCHECKED* items.
             Therefore all-checked is indicated by an empty set
}
var
  clb : TfCheckListBx;
begin
  result := [];
  clb := TfCheckListBx.create(application);
  try
    result := clb.execute(mode,title,msg,itemlist,showbuttons,hlpctxt);
  finally
    clb.free;
  end;
end;


// 
//                                                                              
//    implementation of the dialog's own methods                                
//                                                                              
// 
function TfCheckListBx.Execute(const mode : Tcbxmode;
                               const Title,msg,ItemList : string;
                               const ShowButtons : boolean;
                               const hlpctxt:THelpContext):TcbxResult;
var
  s,x : string;
  QuarterInch,w,htot,i,nob,lwmax,wmax : integer;
begin
  result := [];  // safety default
  QuarterInch := screen.pixelsperinch div 4;

  // mode
  fmode := mode;

  // title
  caption := title;

  // message
  fmsg := msg;

  // help context
  fhlpctxt := hlpctxt;

  // put items into list
  if not assigned(fitems) then fitems := tstringlist.create;  // (freed in close)
  fitems.clear;
  x := itemlist;
  fchecked := [];
  i := 0;
  while x <> '' do begin
    s := chop(x,'|');
    if copy(s,1,1)='+' then begin
      fchecked := fchecked + [i];
      s := copy(s,2,999);
    end;
    fitems.add(s);
    f1stchars := f1stchars+s[1];
    inc(i);
  end;
  // trap too few or too many items
  if fitems.count=0 then begin
    showmessage('No items in list to check');
    exit;
  end;
  if fitems.count-1 > high(Tcbxint) then begin
    showmessage('Too many items in check box');
    exit;
  end;
  fcurrentix := 0;  // first item selected
  f1stChars := uppercase(f1stChars);



  // calculate some dimensions  1-message  2-list
  wmax := screen.pixelsperinch*2;  // 2 inches wide to start with
  htot := 0;
  //----message---  (display message is done with Paint())
  x := fmsg;
  while x <> '' do begin
    s := chop(x,'|');
    if s='' then s := 'Ay';               // vertical space for blank lines
    htot := htot + canvas.textheight(s);
    w := canvas.textwidth(s);
    if w > wmax then wmax := w;
  end;
  fY0 := htot+quarterinch;  // list starts 1/4" after message
  flineHt := Canvas.textheight('Ay');
  //----items---
  htot := fY0+(flineHt*fitems.count);
  lwmax := 0;
  for i := 0 to fitems.count-1 do begin
    w := canvas.textwidth(fitems.strings[i]) + (flineHt*2);   //(allow space for tick)
    if w > lwmax then lwmax := w;
  end;
  htot := htot+quarterinch;  // buttons start start 1/4" after list
  if lwmax > wmax then wmax := lwmax;

  buok.top := htot;
  bucancel.top := htot;
  buhelp.top := htot;
  w := 0;
  if showbuttons or (HlpCtxt<>0) then htot := htot + buok.height + 10;
  if showbuttons then w := buok.width + bucancel.width + quarterinch;
  if hlpCtxt <>0 then w := w + buhelp.width + quarterinch;
  if w > wmax then wmax := w;

  clientwidth := wmax + 40;
  clientheight := htot;
  fx0 := (clientwidth - lwmax) div 2;  // centralise list

  // buttons  - show those required and layout nicely
  buok.visible := showbuttons and (fmode=cbxmPickSome);
  bucancel.visible := showbuttons;
  buhelp.visible := fhlpctxt<>0;

  nob := 0;
  if buok.visible then inc(nob);
  if bucancel.visible then inc(nob);
  if buhelp.visible then inc(nob);
  if nob > 0 then begin
    w := (clientwidth - (buok.width * nob)) div (nob+1);
    i := 0;
    if buok.visible then begin
      i := w;
      buok.left := i;
      i :=  i + buok.width;
    end;
    if bucancel.visible then begin
      i := i + w;
      bucancel.left := i;
      i :=  i + bucancel.width;
    end;
    if buhelp.visible then begin
      i := i + w;
      buhelp.left := i;
    end;
  end;


  // get input and collect results
  if showmodal = mrok then begin
    if fmode=cbxmCheckAll then result := notchecked else result := checked;
  end else begin
    if fmode=cbxmCheckAll then begin
      if result = [] then result := [0];
    end else begin
      result := [];
    end;
  end;

  fitems.free;

end;


procedure TfCheckListBx.BuHelpClick(Sender: TObject);
begin
  Application.helpcontext(fhlpCtxt);
end;


procedure TfCheckListBx.FormPaint(Sender: TObject);
var
  x,s : string;
  h : integer;
  i : tcbxint;
begin
  // print message text
  if fmsg <> '' then begin
    x := fmsg;
    h := 5;
    canvas.font.color := clgreen;
    while x <> '' do begin
      s := chop(x,'|');
      if s <> '' then begin
        canvas.textout(10,h,s);
        h := h + canvas.textheight(s);
      end else begin
        h := h + canvas.textheight('Ay');   // blank lines get vertical space
      end;
    end;
  end;

  // redraw whole list
  for i := 0 to fitems.count-1 do DrawItem(i);

end;

procedure TfCheckListBx.DrawItem(const ix:TCbxint);
{ do all the drawing required for this list item }
var
  x,y : integer;
begin
  with canvas do begin
    // do text
    font.color := clBlack;
    brush.color := clwindow;
    x := fx0+(fLineHt*2);
    y := fy0+(ix*fLineHt);
    textout(x,y,fitems.strings[ix]);
  end;
  Drawtick(ix);
end;

procedure TfCheckListBx.DrawTick(const ix:TCbxint);
{ draw the tick and any highlight marker for the item }
var
  x,y : integer;
  T,R   : Trect;
  tx,ty : integer;
begin
  x := fX0;
  y := fy0+(ix*fLineHt);
  R := rect(x,y,x+(flineht*2),y+flineht);  // where we want to paint  (tick and highlight)

  // clear
  with canvas do begin
    brush.color := clwindow;
    fillrect(R);

    // locate the bitmap to paint
    if ix in fchecked then tx := 0 else tx:=18;
    ty := ord(fmode) * 18;
    T := rect(tx,ty,tx+18,ty+18);

    // draw it
    R := rect(x,y,x+flineht,y+flineht);  // where we want to paint (tick)
    copyrect(R,ticks.canvas,T);

    // possibly draw highlight
    if ix=fcurrentix then begin
      R := rect(x+flineht,y,x+(flineht*2),y+flineht);  // where we want to paint (highlight)
      brush.color := clSilver;
      fillrect(R);
      brush.color := clWindow;
    end;
  end;
end;


function TfCheckListBx.chop(var Mainstring : string; const Chopper : string):string;
var
  silly:boolean;
  p : integer;
begin
  // trap exceptional cases
  silly := false;
  if Mainstring = '' then begin
    result := '';
    silly := true;
  end;
  if Chopper = '' then begin
    result := Mainstring;
    mainstring := '';
    silly := true;
  end;
  if silly then exit;

  p := pos(Chopper,Mainstring);
  if p = 0 then begin       {not found so return rest}
    result := Mainstring;
    mainstring := '';
  end else begin
    result := copy(Mainstring,1,p-1);    {return LH as function result}
    p := p + length(chopper);            {skip chopper}
    if p > length(Mainstring) then Mainstring := ''  {trap terminal chopper}
    else Mainstring := copy(Mainstring,p,length(mainstring)-(p-1));  {lose LH from main string}
  end;
end;

function  TfCheckListBx.Checked:TCbxresult;
{ set of checked items }
begin
  result := fchecked;
end;

function  TfCheckListBx.NotChecked:TCbxresult;
{ set of not checked items }
var
  i : Tcbxint;
begin
  result := [];
  for i := 0 to fitems.count-1 do begin
    if not(i in fchecked) then result := result + [i];
  end;
end;



procedure TfCheckListBx.FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
{ if the mouse is pointing to an item then set the current index }
var
  outside : boolean;
  ix : TcbxInt;
begin
  // see if mouse is in the working area
  outside := false;
  if (x < fx0) or (x>(clientwidth-fx0)) then outside := true;
  if (y < fy0) or (y>Fy0+(fitems.count*flineht)) then outside := true;
  if outside then ix := Noix
             else ix := ((y-fy0)-1) div flineht;
  setcurrentix(ix);
  //if outside then caption := 'O' else caption := 'x';

end;


procedure TfCheckListBx.SetCurrentix(const ix:TcbxInt);
{ to set the current index we may have to remove a highlight marker
  before resetting }
var
  oldix : tcbxint;
begin
  if ix = fcurrentix then exit;  //no change
  oldix := fcurrentix;;
  fcurrentix := ix;
  if oldix <> NoIx then drawtick(oldix);            //  remove highlight
  if fcurrentix <> Noix then drawtick(fcurrentix);  //  set new highlight
end;



procedure TfCheckListBx.Flipcheck;
{ change the checked state of the currently highlighted item
  If PickOne then this will close
  If CheckAll then this will close if all are checked }
var
  waschecked : boolean;
begin
  if fcurrentix = noix then exit;         // nothing selected
  
  // adjust the state of the ticks
  waschecked := fcurrentix in fchecked;
  if waschecked then fchecked := fchecked - [fcurrentix]
                else fchecked := fchecked + [fcurrentix];
  drawtick(fcurrentix);
  application.processmessages;
  
  // do we have to take any action?
  case fmode of
  cbxmPickOne   : if not waschecked then modalresult := mrok;
  cbxmPickSome  : ; // do nothing we can tick/untick as much as we like
  cbxmCheckAll  : if notchecked=[] then modalresult := mrok;
  end;{case}

end;

procedure TfCheckListBx.FormMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  flipcheck;
end;

procedure TfCheckListBx.FormKeyPress(Sender: TObject; var Key: Char);
{ trap space which flips check and moves down to next item
  if we have a printable character then see if it matches
  any of the first letters of the items.  (Go to the first we find)
  If there is only one then (un)tick it
 }
var
  k,fcs:string;
  i,p : integer;
begin
  // space flips tick
  if key = ' ' then begin
    key := #0;
    flipcheck;
    if fcurrentix = noix then begin
      setcurrentix(0);
    end else begin
      i := fcurrentix+1;
      if i > (fitems.count-1) then i := 0;
      setcurrentix(i);
    end;
    exit;
  end;

  // other key may find a unique item
  if key > ' ' then begin
    k := uppercase(key);
    fcs := f1stchars+f1stchars;
    if fcurrentix = noix then i := 0 else i := fcurrentix;
    fcs := copy(fcs,i+2,fitems.count);  // string index starts with next item
    p := pos(k,fcs);
    if p=0 then exit;   // no match - quit
    // work out the index of this match and set it
    i := (i+p) mod fitems.count;
    setcurrentix(i);

    // if this is unambigious then tick it
    fcs := copy(fcs+' ',p+1,99);
    if pos(k,fcs)=0 then flipcheck;
  end;

end;

procedure TfCheckListBx.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
{ navigation and ok/cancel }
var
  i : Tcbxint;
  tot : integer;
begin
  if key=vk_down then begin
    if fcurrentix = noix then begin
      setcurrentix(0);
    end else begin
      i := fcurrentix+1;
      if i > (fitems.count-1) then i := 0;
      setcurrentix(i);
    end;
    key := 0;
  end;

  if key=vk_up then begin
    if fcurrentix = noix then begin
      setcurrentix(fitems.count-1);
    end else begin
      if fcurrentix=0 then i :=fitems.count-1 else i := fcurrentix-1;
      setcurrentix(i);
    end;
    key := 0;
  end;

  if key=vk_escape then begin
    modalresult := mrcancel;
    key := 0;
  end;

  if key=vk_return then begin
    // normally this only applies to pick-some.  But there is a special
    // case where a default has been set for pick-one
    key := 0;
    if fmode=cbxmPickSome then begin
      modalresult := mrOk;
      exit;
    end;
    if fmode=cbxmPickOne then begin
      tot := 0;
      for i := 0 to fitems.count-1 do begin
        if i in fchecked then inc(tot);
      end;
      if tot = 1 then modalresult := mrOk;
    end;
  end;

end;

end.
